==> ikapi-3.1.35.tgz: package/src/BaseEvent/IKTransaction.js <==
//
import IKGlobals from '../IKGlobals';
import Utils  from "../API/IKUtils";
//Database
import IKDBDDB from "../Database/DDB/IKDB";
import IKDBPSQL from "../Database/PSQL/IKDB";
//Cache
import IKCacheRedis from '../Cache/Redis/IKRedis';
//Comps
import IKLogger from "../Logger/IKLogger";
import IKValidation from '../Validation/IKValidation';
import IKPublisher from '../Publisher/IKPublisher';
import IKExecutionTracker from '../Tracker/IKExecutionTracker';
//Request
import IKRequest from '../API/IKRequest';
import IKResponse, { IKBadRequestResponseWithRollback } from '../API/IKResponse';

//
export default class IKTransaction {
  constructor(event, context, config, _isBatch, _retrowErrors) {
    this._event = event;
    this._context = context;
    this._config = config;
    //response behaviour flags
    this._syncReturn = _isBatch;
    this._retrowErrors = _retrowErrors; /* retrow internal errors */
    this._resp = null;
      //When set, this will be called with the response context right before calling the context suceed/fail - useful for IOing the resp for example.
    this.responseProxy = null;
    //
    this.logger = new IKLogger(config.logger, Utils.logLevel(), (context.awsRequestId ? context.awsRequestId : (event.requestContext ? event.requestContext.requestId : 'unknown')));
    this.request = new IKRequest(this._event, this._context, this);
    this.publisher = new IKPublisher(config.publisher);
    this.validator = new IKValidation(config.validation);
    this.db = this._getDBDriver();
    if (config.resourcesTracker) this.tracker = new IKExecutionTracker(config.resourcesTracker, context, this.db);
    this.cache = this._getCacheDriver();
  }

  //Main interface
  async execute(executionFunc) {
    await this._executeLoggerFlush(async ()=>{
      await this._executeDBTransactions(async ()=>{
        return await this._execute(executionFunc);
      });
    });
    if (this._syncReturn) return this._resp;
  }
  //Executions
  async _execute(executionFunc) {
    let executionFailed = true; //failled til we say no!
    //safe execution handler
    try { //Execute
      this.logger.debug("Starting main request code");
      this._resp = await executionFunc(this);
      //Tracker after successful execution (before succeeding the context, which will cause things to die)
      if (this.tracker) await this.tracker.trackExecAndDrain();
      //Answer client
      if (this._resp && this._resp instanceof IKResponse) {
        await this._resp.build(this._context, this, this._syncReturn);
        executionFailed = !!(this._resp.getBody() && this._resp.getBody().rollback);
      } else if (!this._syncReturn) { //sync return can return other than ikresponse
        this._resp = this._getErrorResponse(IKGlobals.ErrorResponseInvalidServerResponse, IKGlobals.ErrorCode_APIError)
        await this._resp.build(this._context, this, this._syncReturn);
        this.logger.error("Invalid response object from main request code.");
      }
    } catch (e) { /*EXECUTION FAIL*/
      this.logger.error('Exception when executing main request code.');
      this.logger.exception(e);
      //retrow?
      if (this._retrowErrors) throw e;
      //envelope exception?
      if (executionFailed) {
        this._resp = this._getErrorResponse(IKGlobals.ErrorResponseUnhandledError, IKGlobals.ErrorCode_APIError);
        await this._resp.build(this._context, this, this._syncReturn);
      }
    } return executionFailed;
  }
  async _executeDBTransactions(safeExecution) {
    try {
      //start DB
      if (this.db) await this.db.connect();
      //start transaction
      if (this.db) await this.db.beginTransaction();
      //
      let executionFailed = await safeExecution();
      //Commit if not failed, rollback if failed
      if (!executionFailed) {
        if (this.db) await this.db.commit();
      } else {
        this.logger.log("Rolling back DB transactions. Main code failed!");
        if (this.db) await this.db.rollback();
      }
      //Cleanup DB after execution
      if (this.db) await this.db.cleanup();
    } catch (e) {
      this.logger.error('Exception when executing DB transactions.');
      this.logger.log(e);
      //retrow?
      if (this._retrowErrors) throw e;
    }
  }
  async _executeLoggerFlush(safeExecution) {
    try {
      await safeExecution();
      await this.logger.flushLogs();
      //Tracker (if not tracked yet -- exception case)
      if (this.tracker) await this.tracker.trackExecAndDrain();
    } catch (e) {
      this.logger.error('Exception when flushing logs.');
      this.logger.exception(e);
      //retrow?
      if (this._retrowErrors) throw e;
    } finally {
      this.logger.debug("Transaction ended");
    }
  }
  /* DB drivers support */
  _getDBDriver() {
    if (this._config.database) {
      if (this._config.database.type == IKGlobals.DBDrivers.PSQL) return new IKDBPSQL(this._config.database, this);
      else if (Object.keys(this._config.database).length > 0) return new IKDBDDB(this._config.database, this);
    } return null;
  }
  /* Cache drivers support */
  _getCacheDriver() {
    if (this._config.cache) {
      if (this._config.cache.type == IKGlobals.CacheDrivers.REDIS) return new IKCacheRedis(this._config.cache, this);
    } return null;
  }
  /* Response support */
  _getErrorResponse(error, code) {
    return IKBadRequestResponseWithRollback(error, code);
  }
}

==> ikapi-3.1.36.tgz: package/src/BaseEvent/IKTransaction.js <==
//
import IKGlobals from '../IKGlobals';
import Utils  from "../API/IKUtils";
//Database
import IKDBDDB from "../Database/DDB/IKDB";
import IKDBPSQL from "../Database/PSQL/IKDB";
//Cache
import IKCacheRedis from '../Cache/Redis/IKRedis';
//Comps
import IKLogger from "../Logger/IKLogger";
import IKValidation from '../Validation/IKValidation';
import IKPublisher from '../Publisher/IKPublisher';
import IKExecutionTracker from '../Tracker/IKExecutionTracker';
//Request
import IKRequest from '../API/IKRequest';
import IKResponse, { IKBadRequestResponseWithRollback } from '../API/IKResponse';

//
export default class IKTransaction {
  constructor(event, context, config, _isBatch, _retrowErrors) {
    this._event = event;
    this._context = context;
    this._config = config;
    //response behaviour flags
    this._syncReturn = _isBatch;
    this._retrowErrors = _retrowErrors; /* retrow internal errors */
    this._resp = null;
      //When set, this will be called with the response context right before calling the context suceed/fail - useful for IOing the resp for example.
    this.responseProxy = null;
    //
    this.logger = new IKLogger(config.logger, Utils.logLevel(), (context.awsRequestId ? context.awsRequestId : (event.requestContext ? event.requestContext.requestId : 'unknown')));
    this.request = new IKRequest(this._event, this._context, this);
    this.publisher = new IKPublisher(config.publisher);
    this.validator = new IKValidation(config.validation);
    this.db = this._getDBDriver();
    if (config.resourcesTracker) this.tracker = new IKExecutionTracker(config.resourcesTracker, context, this.db);
    this.cache = this._getCacheDriver();
  }

  //Main interface
  async execute(executionFunc) {
    await this._executeLoggerFlush(async ()=>{
      await this._executeDBTransactions(async ()=>{
        return await this._execute(executionFunc);
      });
    });
    if (this._syncReturn) return this._resp;
  }
  //Executions
  async _execute(executionFunc) {
    let executionFailed = true; //failled til we say no!
    //safe execution handler
    try { //Execute
      this.logger.debug("Starting main request code");
      this._resp = await executionFunc(this);
      //Tracker after successful execution (before succeeding the context, which will cause things to die)
      if (this.tracker) await this.tracker.trackExecAndDrain();
      //Answer client
      if (this._resp && this._resp instanceof IKResponse) {
        await this._resp.build(this._context, this, this._syncReturn);
        executionFailed = !!(this._resp.getBody() && this._resp.getBody().rollback);
      } else if (this._syncReturn && this._resp) { //sync return can return other than ikresponse
        executionFailed = false;
      } else {
        this._resp = this._getErrorResponse(IKGlobals.ErrorResponseInvalidServerResponse, IKGlobals.ErrorCode_APIError)
        await this._resp.build(this._context, this, this._syncReturn);
        this.logger.error("Invalid response object from main request code.");
      }
    } catch (e) { /*EXECUTION FAIL*/
      this.logger.error('Exception when executing main request code.');
      this.logger.exception(e);
      //retrow?
      if (this._retrowErrors) throw e;
      //envelope exception?
      if (executionFailed) {
        this._resp = this._getErrorResponse(IKGlobals.ErrorResponseUnhandledError, IKGlobals.ErrorCode_APIError);
        await this._resp.build(this._context, this, this._syncReturn);
      }
    } return executionFailed;
  }
  async _executeDBTransactions(safeExecution) {
    try {
      //start DB
      if (this.db) await this.db.connect();
      //start transaction
      if (this.db) await this.db.beginTransaction();
      //
      let executionFailed = await safeExecution();
      //Commit if not failed, rollback if failed
      if (!executionFailed) {
        if (this.db) await this.db.commit();
      } else {
        this.logger.log("Rolling back DB transactions. Main code failed!");
        if (this.db) await this.db.rollback();
      }
      //Cleanup DB after execution
      if (this.db) await this.db.cleanup();
    } catch (e) {
      this.logger.error('Exception when executing DB transactions.');
      this.logger.log(e);
      //retrow?
      if (this._retrowErrors) throw e;
    }
  }
  async _executeLoggerFlush(safeExecution) {
    try {
      await safeExecution();
      await this.logger.flushLogs();
      //Tracker (if not tracked yet -- exception case)
      if (this.tracker) await this.tracker.trackExecAndDrain();
    } catch (e) {
      this.logger.error('Exception when flushing logs.');
      this.logger.exception(e);
      //retrow?
      if (this._retrowErrors) throw e;
    } finally {
      this.logger.debug("Transaction ended");
    }
  }
  /* DB drivers support */
  _getDBDriver() {
    if (this._config.database) {
      if (this._config.database.type == IKGlobals.DBDrivers.PSQL) return new IKDBPSQL(this._config.database, this);
      else if (Object.keys(this._config.database).length > 0) return new IKDBDDB(this._config.database, this);
    } return null;
  }
  /* Cache drivers support */
  _getCacheDriver() {
    if (this._config.cache) {
      if (this._config.cache.type == IKGlobals.CacheDrivers.REDIS) return new IKCacheRedis(this._config.cache, this);
    } return null;
  }
  /* Response support */
  _getErrorResponse(error, code) {
    return IKBadRequestResponseWithRollback(error, code);
  }
}
